from .otBase import BaseTTXConverter


class table_J_S_T_F_(BaseTTXConverter):
    """Justification table

    The ``JSTF`` table contains glyph substitution and positioning
    data used to perform text justification.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/jstf
    """

    pass
