from .otBase import BaseTTXConverter


class table_S_T_A_T_(BaseTTXConverter):
    """Style Attributes table

    The ``STAT`` table records stylistic or typeface-design attributes that
    differentiate the individual fonts within a font family from one another.
    Those attributes can be used to assist users when navigating the style
    variations of a variable font or a family of static fonts.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/stat
    """

    pass
