from .otBase import BaseTTXConverter


class table_V_A_R_C_(BaseTTXConverter):
    """Variable Components table

    The ``VARC`` table contains variation information for composite glyphs.

    See also https://github.com/harfbuzz/boring-expansion-spec/blob/main/VARC.md
    """

    pass
