from .otBase import BaseTTXConverter


class table_G_D_E_F_(BaseTTXConverter):
    """Glyph Definition table

    The ``GDEF`` table stores various glyph properties that are used
    by OpenType Layout.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/gdef
    """

    pass
