from . import asciiTable


class table_T_T_F_A_(asciiTable.asciiTable):
    """ttfautohint parameters table

    The ``TTFA`` table is used by the free-software `ttfautohint` program
    to record the parameters that `ttfautohint` was called with when it
    was used to auto-hint the font.

    See also http://freetype.org/ttfautohint/doc/ttfautohint.html#miscellaneous-1
    """

    pass
